/////////////////////////////////////////////////////////////////////////////////////////////
//
//		ARENA FILE C   Version 0.3
//
//           -------------------
//
//    copyright: (C) 2008 Projet French Arena
//    web site:  http://www.projet-french-arena.org
//
//    By Arkem
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
/////////////////////////////////////////////////////////////////////////////////////////////




/////////////////////////////////////////////////////////////////////////////////////////////
//  HEADERS
/////////////////////////////////////////////////////////////////////////////////////////////

/*  Standard  */
#include   <errno.h>
#include   <stdio.h>
#include   <stdlib.h>
#include   <string.h>
#include   <unistd.h>
#include   <sys/types.h>

/*  Libraries  */
#include   <SDL/SDL.h>

/*  Program  */
#include   "arena.h"
#include   "palette.h"



/////////////////////////////////////////////////////////////////////////////////////////////
//  MISC
/////////////////////////////////////////////////////////////////////////////////////////////

/*  File type  */
enum   { MAIN_ARENA=0, MAIN_BSA=1, MAIN_BMP=2, MAIN_TEXT=3 };

/*  Palette files  */
#define   MAIN_DEF_PAL   3
char   main_palNames[4][13] = {
	{ "CHARSHT.COL" },
	{ "DAYTIME.COL" },
	{ "DREARY.COL"  },
	{ "PAL.COL"     }
};

/*  Modulo  */
#define   MAIN_MODULO_LESS   -1
#define   MAIN_MODULO_MORE    1

/*  Text output  */
#define   MAIN_TEXTOUT_VIEW   0
#define   MAIN_TEXTOUT_FILE   1


/////////////////////////////////////////////////////////////////////////////////////////////
//  GLOBAL VARIABLES
/////////////////////////////////////////////////////////////////////////////////////////////

/*  Misc  */
// Program end
static int                 main_end = 0;
//  Arena Source Data Structure
static struct arena_data   *main_arenaSource = NULL;
//  Arena Destination Data Structure
static struct arena_data   *main_arenaDest   = NULL;
// Alternate palette
static int                 main_altPalette = MAIN_DEF_PAL;

/*  SDL  */
// Display surface
static SDL_Surface   *main_display;
// Display surface
static SDL_Surface   *main_surface;
// Display palette
static SDL_Color     main_palette[256];

/*  Files  */
// Files number
static int    main_filesNumber;
// Files types
static int    main_sourceFileType;
static int    main_destFileType;
// Files paths
static char   *main_sourceFilePath = NULL;
static char   *main_destFilePath = NULL;
static char   *main_tempFilePath = NULL;

/*  Image  */
// Image data
static u_int8_t   *main_imageData = NULL;
// Animation
static u_int8_t   *main_animData = NULL;
// Image dimensions
static int        main_imageWidth;
static int        main_imageHeight;
// Image number
static int        main_imageNumber = 1;

/*  Scrolling  */
// Start line
static int        main_scrollLine = 0;
// Mode
static int        main_scrollMode = 0;

/*  Text  */
// Text data
static char   *main_textData = NULL;
// Text buffer size
static int    main_textSize;



/////////////////////////////////////////////////////////////////////////////////////////////
//  FONCTIONS DECLARATIONS
/////////////////////////////////////////////////////////////////////////////////////////////

/*  Main  */
// File given by first argument
static int   main_BaseFile();
// File given by second argument
static int   main_DestFile();
// View mode if only one argument
static int   main_Display();
// Display animation frame
//static void  main_DrawImage( int, u_int8_t* );
static void  main_DrawImage( u_int8_t*, u_int8_t* );

/*  GFX  */
// Read BMP image
static int   main_ReadBMP();
// Write BMP image
static int   main_WriteBMP();
// Font rendering exemple
static int   main_FontImage();
// Extract image data and infos from Arena struct
static int   main_ArenaToImage();
// Copy image buffer and infos to Arena struct
static int   main_ImageToArena();
// LightMap rendering exemple
static int   main_LightAnim();
// Palette rendering exemple
static int   main_PaletteImage();

/*  Text  */
// Read Text
static int   main_ReadText();
// Write Text
static int   main_WriteText();
// Extract text data and infos from Arena struct
static int   main_ArenaToText();
// Copy text buffer and infos to Arena struct
static int   main_TextToArena();
// Text Output
static int   main_TextOutput( int );

/*  GLOBAL.BSA  */
// GLOBAL.BSA operations
static int   main_GlobalBSA();
// Save a GLOBAL.BSA file to disk
static int   main_GlobalWrite( char* );
//static int main_GlobalRead();

/*  Utilities  */
// Load an alternate image palette
static int   main_LoadAltPalette( char* );
// Modulate image width
static int   main_Modulo();
// Create a new file path by extracting an existing path and add a new file name
static char* main_newFilePath( char*, char* );   // Existing file path, new file name

/*  Management  */
// Options and Arguments management
static int  main_Arguments( int, char** );
// File initialisation
static int  main_FileInit( char*, int* );
// User events management
static void main_SDLEvents();
// Close everything
static void main_CloseAll();



/////////////////////////////////////////////////////////////////////////////////////////////
//  MAIN FONCTION
/////////////////////////////////////////////////////////////////////////////////////////////
int main( int argc, char **argv )
{
	/*  Locals variables  */
	int   i, tick, anim;

	/*  Initialisations  */
	//  Secured Exit
	atexit( main_CloseAll );
	// Parameters and Arguments
	if( main_Arguments(argc,argv) > 0 )
		return( 1 );
	// SDL library
	if( SDL_Init(SDL_INIT_VIDEO) < 0 )
	{
		fprintf( stderr, "\nSDL_Init error: %s\n", SDL_GetError() );
		return( 1 );
	};

	/*  Base file  */
	i = main_BaseFile();
	if( i > 0 )
		return( i );

	/*  Destination file  */
	if( main_filesNumber > 1 ) {
		// Don't want to go further in any case
		return( main_DestFile() );
	};

	/*  Viewer  */
	i = main_Display();
	if( i > 0 )
		return( i );

	/*  Events loop  */
	tick = 0;
	anim = 0;
	i = 0;
	while( main_end != 1 ) {
		// Timer
		if( SDL_GetTicks() > tick ) {
			if( (main_animData!=NULL) && (anim==0) ) {   // Animation
				SDL_LockSurface( main_display );
				main_DrawImage( (main_imageData+(i*main_imageWidth*main_imageHeight)), main_display->pixels );
				SDL_UnlockSurface( main_display );
				SDL_UpdateRect( main_display, 0, 0, 320, 200 );
				i = ( i + 1 ) % main_imageNumber;
			}
			else if( main_scrollMode != 0 ) {   // Scrolling
				if( (main_scrollMode<0) && (main_scrollLine>0) )
					--main_scrollLine;
				else if( (main_imageHeight-main_scrollLine) > 200 )
					++main_scrollLine;
				SDL_LockSurface( main_display );
				main_DrawImage( (main_imageData+(main_scrollLine*main_imageWidth)), (u_int8_t*)main_display->pixels );
				SDL_UnlockSurface( main_display );
				SDL_UpdateRect( main_display, 0, 0, 320, 200 );
			};
			tick = SDL_GetTicks() + 50;   // 1 tick every 1/20s
			anim = ( anim + 1 ) & 3;      // 1 frame every 4 ticks
		};
		// User events
		main_SDLEvents();
	};

	/*  End  */
	return( 0 );
};



/////////////////////////////////////////////////////////////////////////////////////////////
//  MAIN SUB-FONCTIONS DESCRIPTIONS
/////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////
//  main_BaseFile  //
/////////////////////
// File given by first argument
int main_BaseFile()
{
	/*  Choose appropriate action  */
	if( main_sourceFileType == MAIN_ARENA ) {      // Arena
		main_arenaSource = arena_ReadFile( main_sourceFilePath );
		if( main_arenaSource == NULL )
			return( 1 );
		switch( main_arenaSource->type ) {
			case ARENA_ANIM:
				return( main_ArenaToImage() );
			case ARENA_FONT:
				return( main_FontImage() );   // NEED FONT TO TEXT
			case ARENA_IMAGE:
				return( main_ArenaToImage() );
			case ARENA_LIGHTMAP:
				return( main_LightAnim() );
			case ARENA_PALETTE:
				return( main_PaletteImage() );
			case ARENA_TEXT:
				return( main_ArenaToText() );
			default:
				fputs( "Format is not yet supported ...!", stderr );
				return( 1 );
		};
	}
	else if( main_sourceFileType == MAIN_BMP )     // Standard image
		return( main_ReadBMP() );
	else if( main_sourceFileType == MAIN_TEXT )    // Standard text
		return( main_ReadText() );
	else if( main_sourceFileType == MAIN_BSA ) {   // GLOBAL.BSA
		main_GlobalBSA();
		return( 2 );   // Don't want to go further in any case
	}
	else {                                         // ERROR
		fputs( "main_DestFile error: Unknown file type", stderr );
		return( 1 );
	};

	/*  End (just for compiler)  */
	return( 0 );
};


/////////////////////
//  main_DestFile  //
/////////////////////
// File given by second argument
int main_DestFile()
{
	/*  Choose appropriate action  */
	if( main_destFileType == MAIN_ARENA ) {     // Arena
		// Read base structure (Needed to know exact file type and image offsets)
		main_arenaDest = arena_ReadFile( main_destFilePath );
		if( main_arenaDest == NULL )
			return( 1 );
		// Write file
		switch( main_arenaDest->type )   {
			case ARENA_IMAGE:
				return( main_ImageToArena() );
			case ARENA_TEXT:
				return( main_TextToArena() );
			default:
				fputs( "Format is not supported for writing ...!", stderr );
				return( 1 );
		};
	}
	else if( main_destFileType == MAIN_BMP )    // Standard image
		return( main_WriteBMP() );
	else if( main_destFileType == MAIN_TEXT )   // Standard text
		return( main_WriteText() );
	else {                                      // ERROR
		fputs( "main_DestFile error: Unknown file type", stderr );
		return( 1 );
	};

	/*  End (just for compiler)  */
	return( 0 );
};


////////////////////
//  main_Display  //
////////////////////
// View mode if only one argument
int main_Display()
{
	/*  Local variables  */
	// Draw mode 1
	u_int8_t   *row;
	int        i, done;
	// Draw mode 2
//	int        i, x, y;

	/*  Text  */
	if( main_textData != NULL ) {
		main_TextOutput( MAIN_TEXTOUT_VIEW );
		return( 2 );   // Don't want to go further in any case
	};

	/*  Image  */
	if( main_imageData == NULL ) {
		fputs( "Nothing to display ...!", stderr );
		return( 1 );
	};
	// SDL Window
	SDL_WM_SetCaption( "Arena File", "icon.xpm" );
	// SDL Display view
	main_display = SDL_SetVideoMode( 320, 200, 8, (SDL_SWSURFACE|SDL_HWPALETTE) );
	if( main_display == NULL ) {
		fprintf( stderr, "\nSDL_SetVideoMode Error: %s\n", SDL_GetError() );
		return( 1 );
	};
	// SDL palette
	SDL_SetColors( main_display, main_palette, 0, 256);
	// Animation
	if( main_animData != NULL )
		return( 0 );
	// Fill display
	SDL_LockSurface( main_display );
	// Draw mode 1
	row = main_display->pixels;
	done = 0;
	i = 0;
	while( i < main_imageNumber ) {
		main_DrawImage( (main_imageData+(main_imageWidth*main_imageHeight*i)), (row+(main_imageWidth*i)) );
		++i;
		if( ((i+1-done)*main_imageWidth) > 320 ) {
			++row;
			row += ( 320 * (main_imageHeight-1) );
			done += ( 320 / main_imageWidth );
		};
	};
	// Draw mode 2 (nicer, but image can't be higher than 200 pixels)
//	i = 0;
//	y = 0;
//	while( y < (200/main_imageHeight) ) {
//		x = 0;
//		while( x < (320/main_imageWidth) ) {
//			main_DrawImage( (main_imageData+(main_imageWidth*main_imageHeight*i)), (u_int8_t*)main_display->pixels + ((y*main_imageHeight*320)+(x*main_imageWidth)) );
//			++x;
//			++i;
//			if( i == main_imageNumber )
//				break;
//		};
//		++y;
//		if( i == main_imageNumber )
//			break;
//	};
	SDL_UnlockSurface( main_display );
	// Refresh display
	SDL_UpdateRect( main_display, 0, 0, 320, 200 );

	/*  End  */
	return( 0 );
};


//////////////////////
//  main_DrawImage  //
//////////////////////
// Display animation frame
void main_DrawImage( u_int8_t *src, u_int8_t *dest )
{
	/*  Local variables  */
	int        l;

	/*  Fill display  */
	l = 0;
	while( l < main_imageHeight ) {
		if( (dest+main_imageWidth) >= ((u_int8_t*)main_display->pixels+64000) )   // Stay in display
			break;
		memcpy( dest, src, main_imageWidth );
		src += main_imageWidth;
		dest += 320;
		++l;
	};
};



/////////////////////////////////////////////////////////////////////////////////////////////
//  IMAGES FONCTIONS DESCRIPTIONS
/////////////////////////////////////////////////////////////////////////////////////////////

////////////////////
//  main_ReadBMP  //
////////////////////
// Read BMP image
int main_ReadBMP()
{
	/*  Local variables  */
	u_int8_t   *src, *dst;
	int        i;

	/*  Load Image  */
	main_surface = SDL_LoadBMP( main_sourceFilePath );
	if( main_surface == NULL )   {
		fprintf( stderr, "\nError trying to load BMP file: %s\n", SDL_GetError() );
		return( 1 );
	};

	/*  Image dimensions  */
	printf( "\nSource BMP file: %s\n", main_destFilePath );   // Test
	main_imageWidth  = main_surface->w;
	printf( "Width:    %d\n", main_imageWidth );   // Test
	main_imageHeight = main_surface->h;
	printf( "Height:   %d\n", main_imageHeight );   // Test
	printf( "Scanline: %d\n", main_surface->pitch );   // Test

	/*  Image allocation  */
	main_imageData = (u_int8_t*)malloc( main_imageWidth * main_imageHeight );
	if( main_imageData == NULL )   {
		fprintf( stderr, "main failed to allocate image: %s\n", strerror(errno) );
		return( 1 );
	};

	/*  Image initialisation  */
	SDL_LockSurface( main_surface );
	i = 0;
	src = main_surface->pixels;
	dst = main_imageData;
	while( i < main_imageHeight ) {
		memcpy( dst, src, main_imageWidth );
		src += main_surface->pitch;
		dst += main_imageWidth;
		++i;
	};
	SDL_UnlockSurface( main_surface );

	/*  Image palette  */
	memcpy( main_palette, main_surface->format->palette->colors, (256*sizeof(SDL_Color)) );

	/*  Free surface ( for other use )  */
	SDL_FreeSurface( main_surface );
	main_surface = NULL;

	/*  End  */
	main_imageNumber = 1;
	printf( "\nImage %s loaded\n", main_sourceFilePath );   // Test
	return( 0 );
};


/////////////////////
//  main_WriteBMP  //
/////////////////////
// Write BMP image
int main_WriteBMP()
{
	/*  Local variables  */
	u_int8_t   *src, *dst;
	int        i;

	/*  Need an image to save  */
	if( main_imageData == NULL )   {
		fputs( "BMP output file need an image as source ...!", stderr );
		return( 1 );
	};

	/*  Surface allocation  */
	main_surface = SDL_CreateRGBSurface( SDL_SWSURFACE, main_imageWidth, (main_imageHeight*main_imageNumber), 8, 0, 0, 0, 0 );
	if( main_surface == NULL )   {
		fprintf( stderr, "main failed to create surface: %s\n", SDL_GetError() );
		return( 1 );
	};
	printf( "\nDestination BMP file: %s\n", main_destFilePath );   // Test
	printf( "Scanline: %d\n", main_surface->pitch );   // Test

	/*  Image initialisation  */
	SDL_LockSurface( main_surface );
	i = 0;
	src = main_imageData;
	dst = main_surface->pixels;
	while( i < main_imageHeight ) {
		memcpy( dst, src, main_imageWidth );
		src += main_imageWidth;
		dst += main_surface->pitch;
		++i;
	};
	SDL_UnlockSurface( main_surface );

	/*  Image palette  */
	SDL_SetColors( main_surface, main_palette, 0, 256);

	/*  Save Image  */
	if( SDL_SaveBMP(main_surface,main_destFilePath) < 0 )   {
		fprintf( stderr, "\nError trying to save BMP file: %s\n", SDL_GetError() );
		return( 1 );
	};

	/*  End  */
	printf( "Image saved as %s\n", main_destFilePath );   // Test
	return( 0 );
};


//////////////////////
//  main_FontImage  //
//////////////////////
// Font rendering exemple
int main_FontImage()
{
	/*  Locals variables  */
	int                 p, x, y, h;
	struct arena_char   *font;
	u_int8_t            *pixel;

	/*  Image allocation  */
	main_imageData = (u_int8_t*)calloc( 320, 80 );
	if( main_imageData == NULL )   {
		fprintf( stderr, "main_FontImage failed to allocate image: %s\n", strerror(errno) );
		return( 1 );
	};

	/*  Image generation  */
	font = (struct arena_char*)main_arenaSource->data;
	pixel = main_imageData;
	y = 0;
	while( y < 4 )   {
		h = 0;
		while( h < 16 )   {
			x = 0;
			while( x < 20 )   {
				p = 0;
				while( p < 16 )   {
					if( (font[x].lines[h] & ((1<<15)>>p)) > 0 )
						*pixel = 1;
					else
						*pixel = 0;
					++p;
					++pixel;
				};
				++x;
			};
			++h;
		};
		++y;
		font += 20;
	};
	h = 0;
	while( h < 16 )   {
		x = 0;
		while( x < 16 )   {
			p = 0;
			while( p < 16 )   {
				if( (font[x].lines[h] & ((1<<15)>>p)) > 0 )
					*pixel = 1;
				else
					*pixel = 0;
				++p;
				++pixel;
			};
			++x;
		};
		++h;
		pixel += 64;
	};

	/*  Image dimensions  */
	main_imageWidth = 320;
	main_imageHeight = 80;

	/*  Image palette  */
	main_palette[0].r = 255;
	main_palette[0].g = 255;
	main_palette[0].b = 255;
	main_palette[1].r = 0;
	main_palette[1].g = 0;
	main_palette[1].b = 0;

	/*  End  */
	main_imageNumber = 1;
	return( 0 );
};


/////////////////////////
//  main_ArenaToImage  //
/////////////////////////
// Extract image data and infos from Arena struct
int main_ArenaToImage()
{
	/*  Locals variables  */
	int        i, size = main_arenaSource->width * main_arenaSource->height * main_arenaSource->length;

	/*  Image dimensions  */
	main_imageNumber = main_arenaSource->length;
	main_imageWidth  = main_arenaSource->width;
	main_imageHeight = main_arenaSource->height;
	// Several images ?
//	main_imageHeight = main_arenaSource->height / main_imageNumber;
//	printf( "Image dimensions and number: %d x %d x %d\n", main_imageWidth, main_imageHeight, main_imageNumber );   // Test

	/*  Image allocation  */
	main_imageData = (u_int8_t*)malloc( size );
	if( main_imageData == NULL )   {
		fprintf( stderr, "main_LoadImage failed to allocate image: %s\n", strerror(errno) );
		return( 1 );
	};

	/*  Copy Arena data to image buffer  */
	memcpy( main_imageData, main_arenaSource->data, size );


	/*  Image palette  */
	if( main_arenaSource->palette != NULL ) {
		i = 0;
		while( i < 256 )   {
			main_palette[i].r = main_arenaSource->palette[i].red;
			main_palette[i].g = main_arenaSource->palette[i].green;
			main_palette[i].b = main_arenaSource->palette[i].blue;
			++i;
		};
	}
	else {   // Use default Arena palette
		i = 0;
		while( i < 256 )   {
			main_palette[i].r = palette_defPalette[i][0];
			main_palette[i].g = palette_defPalette[i][1];
			main_palette[i].b = palette_defPalette[i][2];
			++i;
		};
	};

	/*  End  */
	// Animation ?
	if( main_arenaSource->type == ARENA_ANIM )
		main_animData = main_imageData;
	return( 0 );
};


/////////////////////////
//  main_ImageToArena  //
/////////////////////////
// Copy image buffer and infos to Arena struct
int main_ImageToArena()
{
	/*  Locals variables  */
	int        i;

	/*  Need an image to save  */
	if( main_imageData == NULL )   {
		fputs( "main_SaveImage error: Need an image to save ...!", stderr );
		return( 1 );
	};

	/*  New dimensions need a realloc for dest data  */
	if( (main_imageWidth!=main_arenaDest->width) || (main_imageHeight!=main_arenaDest->height) ) {
		fprintf( stderr, "New image needs to have same dimensions as the original. Save cancelled\n" );
		return( 1 );
	};

	/*  Copy new data  */
	memcpy( main_arenaDest->data, main_imageData, (main_imageWidth*main_imageHeight) );

	/*  Copy palette  */
	if( main_arenaDest->palette != NULL ) {
		i = 0;
		while( i < 256 )   {
			main_arenaDest->palette[i].red   = main_palette[i].r;
			main_arenaDest->palette[i].green = main_palette[i].g;
			main_arenaDest->palette[i].blue  = main_palette[i].b;
			++i;
		};
	};

	/*  Save  */
	puts( "" );
	if( arena_WriteFile( main_arenaDest, main_destFilePath ) > 0 )
		return( 1 );
	printf( "\nImage saved as %s\n", main_destFilePath );   // Test

	/*  End  */
	return( 0 );
};


//////////////////////
//  main_LightAnim  //
//////////////////////
// LightMap rendering exemple
int   main_LightAnim()
{
	/*  Locals variables  */
	u_int8_t   *pixel;
	int        f, y, x;

	/*  Image allocation  */
	main_imageData = (u_int8_t*)malloc( 320 * 192 * 13 );   // 16 lines of 16 20x12 color rectangles * 13 frames
	if( main_imageData == NULL )   {
		fprintf( stderr, "main_PaletteImage failed to allocate image: %s\n", strerror(errno) );
		return( 1 );
	};
	main_animData = main_imageData;

	/*  Image dimensions  */
	main_imageWidth  = 320;
	main_imageHeight = 192;
	main_imageNumber = 13;

	/*  Image generation  */
	pixel = main_imageData;
	f = 0;
	while( f < 13 )   {
		y = 0;
		while( y < 192 )   {
			x = 0;
			while( x < 320 )   {
				*pixel = *(u_int8_t*)( main_arenaSource->data + (f*256) + ((y/12)*16) + (x/20) );
				++pixel;
				++x;
			};
			++y;
		};
		++f;
	};

	/*  Default palette  */
	x = 0;
	while( x < 256 )   {
		main_palette[x].r = palette_defPalette[x][0];
		main_palette[x].g = palette_defPalette[x][1];
		main_palette[x].b = palette_defPalette[x][2];
		++x;
	};

	/*  End  */
	return( 0 );
};


/////////////////////////
//  main_PaletteImage  //
/////////////////////////
// Palette rendering exemple
int main_PaletteImage()
{
	/*  Locals variables  */
	u_int8_t   *pixel;
	int        y, x;

	/*  Image allocation  */
	main_imageData = (u_int8_t*)malloc( 320 * 192 );   // 16 lines of 16 20x12 color rectangles
	if( main_imageData == NULL )   {
		fprintf( stderr, "main_PaletteImage failed to allocate image: %s\n", strerror(errno) );
		return( 1 );
	};

	/*  Image dimensions  */
	main_imageWidth = 320;
	main_imageHeight = 192;

	/*  Image generation  */
	pixel = main_imageData;
	y = 0;
	while( y < 192 )   {
		x = 0;
		while( x < 320 )   {
			*pixel = ( (y/12) * 16 ) + ( x / 20 );
			++pixel;
			++x;
		};
		++y;
	};

	/*  Image palette  */
	x = 0;
	while( x < 256 )   {
		main_palette[x].r = main_arenaSource->palette[x].red;
		main_palette[x].g = main_arenaSource->palette[x].green;
		main_palette[x].b = main_arenaSource->palette[x].blue;
		++x;
	};

	/*  End  */
	main_imageNumber = 1;
	return( 0 );
};



/////////////////////////////////////////////////////////////////////////////////////////////
//  TEXT FONCTIONS DESCRIPTIONS
/////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////
//  main_ReadText  //
/////////////////////
// Read Text
int main_ReadText()
{
	/*  Local variables  */
	FILE   *file;
	char   *dest;
	char   c;
	int    i, size;

	/*  Open file  */
	file = fopen( main_sourceFilePath, "rb" );
	if( file == NULL ) {
		fprintf( stderr, "main_ReadText couldn't open %s: %s\n", main_sourceFilePath, strerror(errno) );
		return( 1 );
	};

	/*  Get temporary size  */
	fseek( file, 0, SEEK_END );
	main_textSize = ftell( file );
	rewind( file );

	/*  Text allocation  */
	main_textData = (char*)malloc( main_textSize );
	if( main_textData == NULL ) {
		fprintf( stderr, "main_ReadText failed to allocate buffer: %s\n", strerror(errno) );
		return( 1 );
	};

	/*  Read text  */
	size = main_textSize;
	dest = main_textData;
	i = 0;
	while( i < main_textSize ) {
		fread( &c, 1, 1, file );
		if( c == '|' ) {
			*(dest-1) = 0;
			fseek( file, 1, SEEK_CUR );
			size -= 2;
			i += 2;
		}
		else {
			*dest = c;
			++dest;
			++i;
		};
	};

	/*  End  */
	main_textSize = size;
	fclose( file );
	return( 0 );
};


//////////////////////
//  main_WriteText  //
//////////////////////
// Write Text
int main_WriteText()
{
	/*  Local variables  */
	FILE   *file;
	char   c;
	int    i;

	/*  Need a text data  */
	if( main_textData == NULL )   {
		fputs( "main_WriteText error: No text buffer ...!", stderr );
		return( 1 );
	};

	/*  Open file  */
	file = fopen( main_destFilePath, "wb" );
	if( file == NULL ) {
		fprintf( stderr, "main_WriteText couldn't open %s: %s\n", main_destFilePath, strerror(errno) );
		return( 1 );
	};

	/*  Write text  */
	i = 0;
	while( i < main_textSize ) {
		c = main_textData[i];
		if( c == 0 ) {   // Text separator
			c= 0x0D;
			fwrite( &c, 1, 1, file );
			c = '|';
			fwrite( &c, 1, 1, file );
			c= 0x0D;
			fwrite( &c, 1, 1, file );
		}
		else
			fwrite( &c, 1, 1, file );
		++i;
	};

	/*  End  */
	fclose( file );
	printf( "%s successfully written\n", main_destFilePath );   // Test
	return( 0 );
};


////////////////////////
//  main_ArenaToText  //
////////////////////////
// Extract Text data and infos from Arena struct
int main_ArenaToText()
{
	/*  Text size  */
	main_textSize  = main_arenaSource->length;
//	fprintf( stderr, "Text size: %d\n", main_textSize );   // Test

	/*  Text allocation  */
	main_textData = (char*)malloc( main_textSize );
	if( main_textData == NULL ) {
		fprintf( stderr, "main_ArenaToText failed to allocate buffer: %s\n", strerror(errno) );
		return( 1 );
	};

	/*  Copy Arena data to text buffer  */
	memcpy( main_textData, (char*)main_arenaSource->data, main_textSize );

	/*  End  */
	return( 0 );
};


////////////////////////
//  main_TextToArena  //
////////////////////////
// Copy text buffer and infos to Arena struct
int main_TextToArena()
{
	/*  New size ?  */
	fprintf( stderr, "Text size: Original %d, New %d\n", main_arenaDest->length, main_textSize );   // Test
	if( main_textSize > main_arenaDest->length ) {
		main_arenaDest->data = (char*)realloc( main_arenaDest->data, main_textSize );
		if( main_arenaDest->data == NULL ) {
			fprintf( stderr, "main_TextToArena failed to reallocate Arena data: %s\n", strerror(errno) );
			return( 1 );
		};
		main_arenaDest->length = main_textSize;
	};

	/*  Copy text buffer to Arena data  */
	memcpy( (char*)main_arenaDest->data, main_textData, main_textSize );

	/*  Save  */
	puts( "" );
	if( arena_WriteFile( main_arenaDest, main_destFilePath ) > 0 )
		return( 1 );
//	printf( "\nText saved as %s\n", main_destFilePath );   // Test

	/*  End  */
	return( 0 );
};


///////////////////////
//  main_TextOutput  //
///////////////////////
// Texte Output
int main_TextOutput( int mode )
{
	/*  Local variables  */
	FILE   *output;
	char   *line;
	int    i;

	/*  Open destination file  */
	if( mode == MAIN_TEXTOUT_FILE ) {
		output = fopen( main_destFilePath, "wt" );
		if( output == NULL ) {
			fprintf( stderr, "main_TextOutput couldn't open %s for writing: %s\n", main_destFilePath, strerror(errno) );
			return( 1 );
		};
	}
	else
		output = stdout;

	/*  Text init  */
	if( (main_sourceFileType==MAIN_ARENA) && (main_arenaSource->type==ARENA_PALETTE) ) {
		// Write Header
		fprintf( output, "\nARENA palette values\n" );
		fprintf( output, "Extracted from file: %s\n\n", (strrchr (main_sourceFilePath,'/')+1) );
//		fprintf( output, "Extracted from file: %s\n\n", (strrchr (main_sourceFilePath,'\')+1) );   // For windows
		fprintf( output, "The color components are stored in RED, GREEN, BLUE order\n" );
		fprintf( output, "The first color is usually used as a transparent value by the program\n\n" );
		// Generate Palette values
		i = 0;
		while( i < 256 ) {
			fprintf( output, "%d, %d, %d\n", main_arenaSource->palette[i].red, main_arenaSource->palette[i].green, main_arenaSource->palette[i].blue );
			++i;
		};
	}
	else {
		// Need a text data
		if( main_textData == NULL )   {
			fputs( "main_TextOutput error: No text buffer ...!", stderr );
			return( 1 );
		};
		// Write text
		i = 0;
		while( i < main_textSize ) {
			line = main_textData + i;
			fputs( line, output );
			i += strlen( line );
			if( main_textData[i] == 0 )   // Text separator
				fprintf( output, "\n\n" );    // For readability only
			++i;
		};
	};

	/*  End  */
	if( mode == MAIN_TEXTOUT_FILE ) {
		fclose( output );
		printf( "%s successfully written\n", main_destFilePath );   // Test
	};
	return( 0 );
};



/////////////////////////////////////////////////////////////////////////////////////////////
//  GLOBAL.BSA FONCTIONS DESCRIPTIONS
/////////////////////////////////////////////////////////////////////////////////////////////

//////////////////////
//  main_GlobalBSA  //
//////////////////////
// GLOBAL.BSA operations
int main_GlobalBSA()
{
	/*  Local variables  */
	char   **list;
	char   *ext, *name = NULL;
	int    i;

	/*  Extract path end  */
	if( main_filesNumber == 2 ) {   // Only for files write
		i = strlen( main_destFilePath );
		while( i > 0 )   {
			if( main_destFilePath[i] == '/' ) {
//			if( main_destFilePath[i] == '\' ) {   // Windows version
				name = &main_destFilePath[i+1];
				break;
			}
			else
				--i;
		};
		if( i == 0 )
			name = main_destFilePath;   // File name isn't preceeded py a path name
	};

	// Write file
	if( (name != NULL) && (name[0]!='*') ) {   // Extract one file
		if( main_GlobalWrite(name) > 0 )
			return( 1 );
	}
	else {
		// Get GLOBAL.BSA files list
		list = arena_GlobalList( main_sourceFilePath );
		if( list == NULL )
			return( 1 );
		// Per extension mode
		if( (name != NULL) && (name[1]=='.') )
			ext = &name[2];
		// Create directories
//		...
		// Extract / display files
		i = 0;
		while( i < arena_globalNumber ) {
			if( name == NULL )         // Display list mode
				printf( "%s\n", list[i] );
			else if( ext == NULL ) {   // All files extraction mode
//				printf( "Extracting: %s...\n", list[i] );   // Test
				if( main_GlobalWrite(list[i]) > 0 )
					return( 1 );
			}
			else {                     // Per extension extraction mode
				if( strcmp( &list[i][strlen(list[i])-3], ext ) == 0 ) {
//					printf( "Extracting: %s...\n", list[i] );   // Test
					if( main_GlobalWrite(list[i]) > 0 )
						return( 1 );
				};
			};
			++i;
		};
		return( 0 );
	};

	/*  End  */
	return( 0 );
};


////////////////////////
//  main_GlobalWrite  //
////////////////////////
// Save a GLOBAL.BSA file to disk
int main_GlobalWrite( char *name )
{
	/*  Local variables  */
	FILE   *file;
	char   *filePath;

	/*  Load file  */
	main_arenaSource = arena_GetGlobalFile( main_sourceFilePath, name );
	if( main_arenaSource == NULL )
		return( 1 );

	/*  Generate destination file path  */
	filePath = main_newFilePath( main_destFilePath, name );
	if( filePath == 0 ) {
		arena_FreeData( main_arenaSource );
		return( 1 );
	};

	/*  Open destination  */
	file = fopen( filePath, "wb" );
	if( file == NULL ) {
		fprintf( stderr, "main_GlobalWrite couldn't open %s for writing: %s\n", filePath, strerror(errno) );
		arena_FreeData( main_arenaSource );
		return( 1 );
	};

	/*  Write destination  */
	if( fwrite(main_arenaSource->data,1,main_arenaSource->length,file) < main_arenaSource->length ) {
		fprintf( stderr, "main_GlobalWrite: Error writing %s: %s\n", filePath, strerror(errno) );
		arena_FreeData( main_arenaSource );
		fclose( file );
		return( 1 );
	};

	/*  End  */
	arena_FreeData( main_arenaSource );
	fclose( file );
	printf( "%s Writed to %s \n", name, filePath );   // Test
	return( 0 );
};



/////////////////////////////////////////////////////////////////////////////////////////////
//  UTILITIES FONCTIONS DESCRIPTIONS
/////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////
//  main_LoadAltPalette  //
///////////////////////////
// Load an alternate image palette
int main_LoadAltPalette( char *name )
{
	/*  Locals variables  */
	struct arena_data   *data = NULL;
	char                *palPath;
	int                 i;

	/*  Create file path  */
	palPath = main_newFilePath( main_sourceFilePath, name );
	if( palPath == NULL )
		return( 1 );
//	printf( "New palette file path: %s\n", palPath );   // Test

	/*  Read palette  */
	data = arena_ReadFile( palPath );
	if( data == NULL )
		return( 1 );

	/*  Init display palette  */
	i = 0;
	while( i < 256 ) {
		main_palette[i].r = data->palette[i].red;
		main_palette[i].g = data->palette[i].green;
		main_palette[i].b = data->palette[i].blue;
		++i;
	};

	/*  End  */
	arena_FreeData( data );
	return( 0 );
};


///////////////////
//  main_Modulo  //
///////////////////
// Modulate image width
int main_Modulo( int mod )
{
	/*  Locals variables  */
	SDL_Rect   rect;
	int        i, size = main_imageWidth * main_imageHeight;

	/*  Avoid animations and multiple images  */
	if( main_imageNumber > 1 )
		return( 0 );

	/*  Search good values  */
	i = 1;
	while( (size%(main_imageWidth+(i*mod))) > 0 ) {
//		printf( "Modulo: %d / %d = %d : %d\n", size, (main_imageWidth+(i*mod)), (size/(main_imageWidth+(i*mod))), (size%(main_imageWidth+(i*mod))) );
		++i;
		// Avoid division by zero (DOESN'T WORK)
		if( ((main_imageWidth+(i*mod))) == 0 )
			return( 1 );
	};

	/*  Clear old display rectangle  */
	rect.x = 0;
	rect.y = 0;
	rect.w = main_imageWidth;
	rect.h = main_imageHeight;
	SDL_FillRect( main_display, &rect, 0 );

	/*  Image dimensions  */
	main_imageWidth  = main_imageWidth + ( i * mod );
	main_imageHeight = size / main_imageWidth;
	printf( "New dimensions: %d x %d\n", main_imageWidth, main_imageHeight );

	/*  Update Display  */
	// Fill display
	SDL_LockSurface( main_display );
	main_DrawImage( main_imageData, main_display->pixels );
	SDL_UnlockSurface( main_display );
	// Refresh display
	SDL_UpdateRect( main_display, 0, 0, 320, 200 );

	/*  End  */
	return( 0 );
};


////////////////////////
//  main_newFilePath  //
////////////////////////
// Create a new file path by extracting an existing path and add a new file name
char* main_newFilePath( char *filePath, char *file )
{
	/*  Locals variables  */
	int           i, end;

	/*  File path lenght  */
	end = 0;
	i = 0;
	while( i < strlen(filePath) ) {
		if( filePath[i] == '/' )
//		if( filePath[i] == '\' )   // Windows version
			end = i;
		++i;
	};

	/*  Allocate new file path  */
	if( main_tempFilePath != NULL ) {
		main_tempFilePath = realloc( main_tempFilePath, (end+1+strlen(file)+1) );   // end + '/' + file + '\0'
		if( main_tempFilePath == NULL ) {
			fprintf( stderr, "main_newFilePath failed to allocate new file path for %s: %s\n", file, strerror(errno) );
			return( NULL );
		};
	}
	else {
		main_tempFilePath = malloc( end + 1 + strlen(file) + 1 );   // end + '/' + file + '\0'
		if( main_tempFilePath == NULL ) {
			fprintf( stderr, "main_newFilePath failed to allocate new file path for %s: %s\n", file, strerror(errno) );
			return( NULL );
		};
	};

	/*  Copy path and file  */
	if( end > 0 ) {   // filePath had no path
		strncpy( main_tempFilePath, filePath, (end+1) );
		strcpy( &main_tempFilePath[end+1], file );
	}
	else
		strcpy( main_tempFilePath, file );

	/*  End  */
	return( main_tempFilePath );
};



/////////////////////////////////////////////////////////////////////////////////////////////
//  MANAGEMENT FONCTIONS DESCRIPTIONS
/////////////////////////////////////////////////////////////////////////////////////////////

//////////////////////
//  main_Arguments  //
//////////////////////
// Parameters and Arguments management
int main_Arguments( int argc, char **argv )
{
	/*  Locals variables  */
	int           i;
	static char   optstring[] = "h";

	/*  Options  */
	while( (i = getopt(argc,argv,optstring)) != -1 ) {
		switch( i ) {
			case 'h':
				puts( "\n\tArena File help" );
				puts( "\nUsage:" );
				puts( " arenafile [FILE] - View file if supported" );
				puts( " arenafile [ARENA FILE] [STD FILE] - Convert Arena file to standard format file" );
				puts( " arenafile [STD FILE] [ARENA FILE] - Update Arena file with standard format file" );
				puts( " arenafile [PATH TO]GLOBAL.BSA - Display GLOBAL.BSA files list" );
				puts( " arenafile [PATH TO]GLOBAL.BSA [ARENA FILE] - Extract file from GLOBAL.BSA" );
				puts( " arenafile [PATH TO]GLOBAL.BSA *.[EXT] - Extract all [EXT] files from GLOBAL.BSA" );
				puts( " arenafile [PATH TO]GLOBAL.BSA * - Extract all files from GLOBAL.BSA" );
				puts( "\nIt is possible to use a source path with a non-existing Arena file. The" );
				puts( "program will try to get the file in GLOBAL.BSA if present in the same directory." );
//				puts( "Destination files will be written separately, not in GLOBAL.BSA." );
				puts( "\nPalette values can be saved as text." );
				puts( "\nDAT text files are extracted with a special | character used to separate" );
				puts( "text chunks. Don't remove it if you plan to convert the file back to DAT." );
//				puts( "\nGLOBAL.BSA multiple files extraction operations create directories named by" );
//				puts( "file extensions to avoid directory flood." );
				puts( "\nGLOBAL.BSA operations destination can also have a path to avoid extracting files" );
				puts( "to current directory." );
				puts( "GLOBAL.BSA file extensions (EXT) are:" );
				puts( "CFA, CIF, DFA, IMG, INF, MIF, RMD, SET, VOC, XFM and XMI." );
				puts( "\nSupported Arena formats:" );
				puts( " IMG images (non-compressed read and write)" );
				puts( " RCI animations (view only)" );
				puts( " DAT and INF texts (read and write)" );
				puts( " COL palettes and DAT fonts (read only)" );
				puts( "\nSupported standard formats:" );
				puts( " BMP images (read and write)" );
				puts( " Text (write only)" );
				puts( "\nOptions:" );
				puts( " -h   This help" );
				puts( "\nViewer mode keys:" );
				puts( " Esc                     Quit" );
				puts( " Up and Down Arrows      Scroll up and down big images" );
				puts( " Alt-return              Toogle FullScreen / Window mode" );
//				puts( " s                       Save displayed image in current directory" );
				puts( " Tab                     Toogle palettes if possible" );
				puts( " Left and Right Arrows   Modulate display lines size" );
				return( 1 );
				break;
			default:
				fprintf( stderr, "\nUnknown option\n" );
		};
	};

	/*  Argument 1 - FileName  */
	// Need at least one
	if( optind >= argc ) {
		fprintf( stderr, "\nThis program need at least one FileName as arguments...\n" );
		return( 1 );
	};
	// Generate appropriate path and file name
	main_filesNumber = 1;
	if( main_FileInit(argv[optind],&main_sourceFileType) > 0 )
		return( 1 );
	// Index
	++optind;

	/*  Argument 2 - Destination name  */
	// Generate appropriate path and file name
	if( optind < argc ) {
		main_filesNumber = 2;
		if( main_FileInit(argv[optind],&main_destFileType) > 0 )
			return( 1 );
		// Index
		++optind;
	};

	/*  Argument 3 - Seems that an error occured  */
	while( optind < argc ) {
		fprintf( stderr, "\nToo many arguments: %s rejected.\n", argv[optind] );
		++optind;
	};

	/*  End  */
	return( 0 );
};


/////////////////////
//  main_FileInit  //
/////////////////////
// File initialisation
int main_FileInit( char *filePath, int *type )
{
	/*  Local variables  */
	int    lenght = strlen( filePath );
	char   *fp, *ext;

	/*  Check extension  */
	if( (filePath[lenght-1]=='*') || (filePath[lenght-4]=='*') )
		*type = MAIN_BSA;
	else {
		ext = &filePath[lenght-3];
		if( strcmp("bmp",ext) == 0 )
			*type = MAIN_BMP;
		else if( strcmp("BSA",ext) == 0 )
			*type = MAIN_BSA;
		else if( strcmp("CEL",ext) == 0 )
			*type = MAIN_ARENA;
//		else if( strcmp("CFA",ext) == 0 )
//			*type = MAIN_ARENA;
		else if( strcmp("CIF",ext) == 0 )
			*type = MAIN_ARENA;
		else if( strcmp("COL",ext) == 0 )
			*type = MAIN_ARENA;
		else if( strcmp("CLR",ext) == 0 )
			*type = MAIN_ARENA;
		else if( strcmp("DAT",ext) == 0 )
			*type = MAIN_ARENA;
//		else if( strcmp("DFA",ext) == 0 )
//			*type = MAIN_ARENA;
		else if( strcmp("IMG",ext) == 0 )
			*type = MAIN_ARENA;
		else if( strcmp("INF",ext) == 0 )
			*type = MAIN_ARENA;
		else if( strcmp("LGT",ext) == 0 )
			*type = MAIN_ARENA;
		else if( strcmp("MIF",ext) == 0 )
			*type = MAIN_ARENA;
		else if( strcmp("NTR",ext) == 0 )
			*type = MAIN_ARENA;
		else if( strcmp("SET",ext) == 0 )
			*type = MAIN_ARENA;
		else if( strcmp("RCI",ext) == 0 )
			*type = MAIN_ARENA;
//		else if( strcmp("RMD",ext) == 0 )
//			*type = MAIN_ARENA;
		else if( strcmp("TXT",ext) == 0 )   // Must be uppercase to be reconized as an Arena File (can be text or texture)
			*type = MAIN_ARENA;
		else if( strcmp("txt",ext) == 0 )   // Must be lowercase to be reconized as a standard text file
			*type = MAIN_TEXT;
//		else if( strcmp("VOC",ext) == 0 )
//			*type = MAIN_ARENA;
//		else if( strcmp("XFM",ext) == 0 )
//			*type = MAIN_ARENA;
//		else if( strcmp("XMI",ext) == 0 )
//			*type = MAIN_ARENA;
		else	{
			fprintf( stderr, "Format of file %d is unknown ...!\n", main_filesNumber );
			return( 1 );
		};
	};

	/*  Generate Path and File names  */
	// PathName allocation
	fp = malloc( lenght + 1 );
	// Copying PathName
	strcpy( fp, filePath );
	// Source or destination ?
	if( main_filesNumber > 1 )
		main_destFilePath = fp;
	else
		main_sourceFilePath = fp;
//	printf( "File path: %s\n", fp );   // Test

	/*  End  */
	return( 0 );
};


//////////////////////
//  main_SDLEvents  //
//////////////////////
// User events management
void main_SDLEvents()
{
	/*  Local variables  */
	int         alt = 0;
	SDL_Event   event;

	/*  Events  */
	while( SDL_PollEvent(&event) )
	{
		switch( event.type )
		{
			case SDL_KEYDOWN:
				switch( event.key.keysym.sym )
				{
					case SDLK_LEFT:    // Modulo - Line decrease
						main_Modulo( MAIN_MODULO_LESS );
						break;
					case SDLK_RIGHT:   // Modulo - Line increase
						main_Modulo( MAIN_MODULO_MORE );
						break;
					case SDLK_UP:    // Scroll Up (for high single images)
						if( main_imageNumber == 1 )
							main_scrollMode = -1;
						break;
					case SDLK_DOWN:    // Scroll Down (for high single images)
						if( main_imageNumber == 1 )
							main_scrollMode = 1;
						break;
					case SDLK_LALT:    // Alter
						alt = 1;
						break;
					case SDLK_ESCAPE:  // Quit
						main_end = 1;
						break;
					case SDLK_TAB:     // Change palette if possible
						main_altPalette = ( main_altPalette + 1 ) & 3;
//						printf( "Alternate palette: %d - %s\n", main_altPalette, main_palNames[main_altPalette] );   // Test
						if( main_LoadAltPalette(main_palNames[main_altPalette]) == 0 )
							SDL_SetColors( main_display, main_palette, 0, 256);
						break;
					default:
						break;
				};
				break;
			case SDL_KEYUP:
				switch( event.key.keysym.sym )
				{
					case SDLK_UP:     // Stop Scrolling Up
						if( main_scrollMode == -1 )
							main_scrollMode = 0;
						break;
					case SDLK_DOWN:   // Stop Scrolling Down
						if( main_scrollMode == 1 )
							main_scrollMode = 0;
						break;
					case SDLK_RETURN: // Toggle FullScreen
						SDL_WM_ToggleFullScreen( main_display );
						break;
					default:
						break;
				};
				break;
			case SDL_QUIT:
				main_end = 1;
				break;
			default:
				break;
		};
	};
};


/////////////////////
//  main_CloseAll  //
/////////////////////
// Close everything
void main_CloseAll()
{
	/*  Console view  */
	puts( "" );

	/*  Files and paths strings memory allocations  */
	if( main_sourceFilePath != NULL )
		free( main_sourceFilePath );
	if( main_destFilePath != NULL )
		free( main_destFilePath );
	if( main_tempFilePath != NULL )
		free( main_tempFilePath );

	/*  Arena Data  */
	arena_FreeAll();

	/*  Image  */
	if( main_imageData != NULL )
		free( main_imageData );

	/*  Text  */
	if( main_textData != NULL )
		free( main_textData );

	/*  SDL  */
	SDL_Quit();
};



/////////////////////////////////////////////////////////////////////////////////////////////
//  END
/////////////////////////////////////////////////////////////////////////////////////////////
